/*!
        @file    $Id:: field.h #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru) 
                 $LastChangedBy: matufuru $

        @date    $LastChangedDate:: 2013-10-20 19:28:34 #$

        @version $LastChangedRevision: 972 $
*/

#ifndef FIELD_INCLUDED
#define FIELD_INCLUDED

#include <valarray>
#include <string>
#include <assert.h>

#include "commonParameters.h"
#include "communicator.h"
#include "bridge_complex.h"

//#define USE_EXPR_TEMPL

//! Container of Field-type object.
/*!
   This class defines field-type quantity which has three
   size parameters, Nin: on-site degree of freedom,
   Nvol: site d.o.f, Nex: extra d.o.f.
   The detailed structure of these degrees of freedom is not
   defined in this class but in subclasses.
   Expression template was implemented by J.Noaki.
                                   [28 Dec 2011 H.Matsufuru]
   Expression template is no longer used. [7 Jun  2014 H.M.]
 */
class Field{

  // friend dcomplex ddotc_complex(Field&, Field&);

 public:
  enum complexness {COMPLEX, REAL};

 protected:
  std::valarray<double> field;
  int m_Nvol;  // lattice volume
  int m_Nin;   // internal d.o.f.
  int m_Nex;   // external d.o.f.
  complexness m_complexness;
  // total degree of freedom is Nin * Nsite * Nex.
  int m_Ntot;  // total d.o.f.
  
  int myindex(const int jin, const int site, const int jex)
    const {  return jin + m_Nin*(site + m_Nvol*jex); }

  Bridge::VerboseLevel m_vl;

public:

  Field():
  m_Nvol(0), m_Nin(0), m_Nex(0), m_complexness(COMPLEX),
  m_vl(CommonParameters::Vlevel()) {  }

 Field(const int Nin, const int Nvol, const int Nex,
                                  const complexness cmpl = COMPLEX):
  m_Nvol(Nvol), m_Nin(Nin), m_Nex(Nex), m_complexness(cmpl),
  m_vl(CommonParameters::Vlevel()) {
    m_Ntot = m_Nin * m_Nvol * m_Nex;
    field.resize(m_Ntot);
  }

  void reset(const int Nin, const int Nvol, const int Nex,
             const complexness cmpl = COMPLEX){
    m_Nin  = Nin;
    m_Nvol = Nvol;
    m_Nex  = Nex;
    m_Ntot = m_Nin * m_Nvol * m_Nex;
    m_complexness = cmpl;
    field.resize(m_Ntot);
  }

  int nin() const { return m_Nin;}
  int nex() const { return m_Nex;}
  int nvol() const { return m_Nvol;}
  int ntot() const { return m_Ntot;}
  int size() const { return m_Nin * m_Nvol * m_Nex;}
  complexness field_complexness() const { return m_complexness;}

  double cmp(const int jin, const int site, const int jex) const {
    return field[myindex(jin,site,jex)];
  }

  double cmp(const int i) const { return field[i]; }

  double* ptr(const int i){ return &field[i]; }

  void set(const int jin, const int site, const int jex, double v){
    field[myindex(jin,site,jex)] = v;
  }

  void set(const int i, double v){ field[i] = v; }

  void add(const int jin, const int site, const int jex, double v){
    field[myindex(jin,site,jex)] += v;
  }

  void add(const int i, double v){ field[i] += v; }

  void setpart_ex(int ex, const Field& w, int exw);
  void addpart_ex(int ex, const Field& w, int exw);
  void addpart_ex(int ex, const Field& w, int exw, double prf);

  double norm() const {
    double b = norm2();
    return sqrt(b);
  }

  // following several function was added to test the
  // performance of the functions replacing expression
  // templates.  [8 Apr 2012 H.Matsufuru]

  dcomplex zdotc(const Field& x) const;

  double ddotc(const Field& x) const;

  double ddotc_ex(const int ex1, const Field& x, const int ex2) const;

  void ddotc(double& yx, const Field& x);

  double norm2() const;

  void norm2(double& yy);

  //! y = a*x + w.
  void daxpy_copy(double   a, const Field& x, const Field& w);

  void zaxpy_copy(dcomplex a, const Field& x, const Field& w);

  void daxpy(double   a, const Field& x);

  void zaxpy(dcomplex a, const Field& x);

  void dscal(double a);

  void dcopy(const Field& x);

  void dcopy(double a, const Field& x);

  void clear();

 protected:
  // functions specific to complex case.
  void zaxpy_cmplx(dcomplex a, const Field& x);
  void zaxpy_copy_cmplx(dcomplex a, const Field& x, const Field& w);
  dcomplex zdotc_cmplx(const Field& x) const;

 public:

  //! \brief determines the statistics of the field.
  //! average, maximum value, and deviation is determined
  //! over global lattice. On-site degree of freedom is
  //! sumed over in quadrature, not averaged.
  //! This function works only on single node.
  void stat(double& Fave, double& Fmax, double& Fdev);

  //! write field values to a text file.
  void write_text(std::string);

  //! \brief read field values from text file.
  //! Assumes field size is already defined, and if it is
  //! inconsistent with the field in the file, aborted.
  //! This function works only on single node.
  void read_text(std::string);

  Field& operator-();
  Field& operator=(const double&);
  Field& operator+=(const Field&);
  Field& operator-=(const Field&);
  Field& operator*=(const double&);
  Field& operator/=(const double&);
  double operator*(const Field& rhs);

};

inline Field& Field::operator-(){
  field = -field;
  return *this;
}
inline Field& Field::operator=(const double& r){
  field = r;
  return *this;
}
inline Field& Field::operator+=(const Field& rhs){
  field += rhs.field;
  return *this;
}
inline Field& Field::operator-=(const Field& rhs){
  field -= rhs.field;
  return *this;
}
inline Field& Field::operator*=(const double& rhs){
  field *= rhs;
  return *this;
}

inline Field& Field::operator/=(const double& rhs){
  field /= rhs;
  return *this;
}

inline double Field::operator*(const Field& rhs){
    double a = (field*rhs.field).sum();
    double b = Communicator::reduce_sum(a);
    return b;
}

inline dcomplex ddotc_complex(const Field& y, const Field& x){
  return y.zdotc(x);
}

inline dcomplex zdotc(const Field& y, const Field& x){
  return y.zdotc(x);
}

//----------------------------------------------------------------

inline int exchange (int count, Field* recv_buf, Field* send_buf, int idir, int ipm, int tag)
{
  return Communicator::exchange(count, recv_buf->ptr(0), send_buf->ptr(0), idir, ipm, tag);
}

inline int send_1to1 (int count, Field* recv_buf, Field* send_buf, int p_to, int p_from, int tag)
{
  return Communicator::send_1to1(count, recv_buf->ptr(0), send_buf->ptr(0), p_to, p_from, tag);
}

//----------------------------------------------------------------

inline Field operator* (const Field& v, const double s)
{
  Field w(v);
  w *= s;
  return w;
}

inline Field operator* (const double s, const Field& v)
{
  return operator*(v,s);
}

inline Field operator+ (const Field& lhs, const Field& rhs)
{
  Field w(lhs);
  w += rhs;
  return w;
}

inline Field operator- (const Field& lhs, const Field& rhs)
{
  Field w(lhs);
  w -= rhs;
  return w;
}

#endif
