/*!
        @file    $Id:: fopr_Domainwall.cpp #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru)
                 $LastChangedBy: matufuru $

        @date    $LastChangedDate:: 2014-06-02 23:54:18 #$

        @version $LastChangedRevision: 1054 $
*/

#include "fopr_Domainwall.h"

using std::valarray;
using std::string;

const std::string Fopr_Domainwall::class_name = "Fopr_Domainwall";

//====================================================================
void Fopr_Domainwall::init()
{
  int Nc   = CommonParameters::Nc();
  int Nd   = CommonParameters::Nd();
  m_NinF = 2 * Nc * Nd;
  m_Nvol = CommonParameters::Nvol();

  m_v4.reset(m_NinF, m_Nvol, 1);
  m_w4.reset(m_NinF, m_Nvol, 1);
  m_w4up.reset(m_NinF, m_Nvol, 1);
  m_w4dn.reset(m_NinF, m_Nvol, 1);
}

//====================================================================
void Fopr_Domainwall::set_parameters(const double mq, const double M0,
                                     const int Ns, const valarray<int> bc)
{
  int Ndim = CommonParameters::Ndim();

  //- print input parameters
  vout.general(m_vl, "Parameters of %s:\n", class_name.c_str());
  vout.general(m_vl, "  mq   = %8.4f\n", mq);
  vout.general(m_vl, "  M0   = %8.4f\n", M0);
  vout.general(m_vl, "  Ns   = %4d\n", Ns);
  for (int mu = 0; mu < Ndim; ++mu) {
    vout.general(m_vl, "  boundary[%d] = %2d\n", mu, bc[mu]);
  }

  assert(bc.size() == Ndim);

  //- store values
  m_mq = mq;
  m_M0 = M0;
  m_Ns = Ns;

  m_boundary.resize(Ndim);
  for (int mu = 0; mu < Ndim; ++mu) {
    m_boundary[mu] = bc[mu];
  }

  //- propagate parameters
  double kappa = 1.0 / (8.0 - 2.0 * m_M0);
  m_fopr_w->set_parameters(kappa, m_boundary);

  m_w1.reset(m_NinF, m_Nvol, m_Ns);
  m_w2.reset(m_NinF, m_Nvol, m_Ns);
  m_v1.reset(m_NinF, m_Nvol, m_Ns);
  m_v2.reset(m_NinF, m_Nvol, m_Ns);

}

//====================================================================
void Fopr_Domainwall::DdagD(Field& v, const Field& w)
{
  D(m_w1,w);
  mult_gm5R(m_w2,m_w1);
  D(m_w1,m_w2);
  mult_gm5R(v,m_w1);
}

//====================================================================
void Fopr_Domainwall::Ddag(Field& v, const Field& w)
{
  mult_gm5R(m_w2,w);
  D(m_w1,m_w2);
  mult_gm5R(v,m_w1);
}

//====================================================================
void Fopr_Domainwall::H(Field& v, const Field& w)
{
  D(m_w1,w);
  mult_gm5R(v,m_w1);
}

//====================================================================
void Fopr_Domainwall::mult_chproj_4d(Field& v, const Field& w, int ipm)
{

  double Fch;
  if (ipm == 1) {
    Fch =  1.0;
  } else if (ipm == -1) {
    Fch = -1.0;
  } else {
    vout.crucial(m_vl, "%s: illegal chirality = %d.\n",
                                          class_name.c_str(), ipm);
    abort();
  }

  v.dcopy(w);  // v = w;
  m_fopr_w->mult_gm5(m_w4,w);
  v.daxpy(Fch, m_w4);
  v.dscal(0.5);  // v *= 0.5;
}

//====================================================================
void Fopr_Domainwall::mult_gm5R(Field& v, const Field& w)
{

  assert(w.nin() == m_NinF);
  assert(w.nvol() == m_Nvol);
  assert(w.nex() == m_Ns);
  assert(v.nin() == m_NinF);
  assert(v.nvol() == m_Nvol);
  assert(v.nex() == m_Ns);

  #pragma omp barrier

  for (int is = 0; is < m_Ns; ++is) {
    m_w4.setpart_ex(0, w, is);
    m_fopr_w->mult_gm5(m_v4, m_w4);
    v.setpart_ex(m_Ns - 1 - is, m_v4, 0);
  }

  #pragma omp barrier

}
//====================================================================
void Fopr_Domainwall::D(Field& v, const Field& w)
{

  assert(w.nin() == m_NinF);
  assert(w.nvol() == m_Nvol);
  assert(w.nex() == m_Ns);
  assert(v.nin() == m_NinF);
  assert(v.nvol() == m_Nvol);
  assert(v.nex() == m_Ns);

  double Fmid = 4.0 - m_M0;

  #pragma omp barrier

  for (int is = 0; is < m_Ns; ++is) {
    v.setpart_ex(is, w, is);
    m_fopr_w->D_ex(m_v4, 0, w, is);
    v.addpart_ex(is, m_v4, 0, Fmid);

    int    is_up = (is + 1) % m_Ns;
    double Fup   = -0.5;
    if (is == m_Ns - 1) Fup = m_mq / 2.0;
    m_w4up.setpart_ex(0, w, is_up);
    m_fopr_w->mult_gm5(m_v4, m_w4up);
    m_w4up.daxpy(-1.0, m_v4);
    v.addpart_ex(is, m_w4up, 0, Fup);

    int    is_dn = (is - 1 + m_Ns) % m_Ns;
    double Fdn   = -0.5;
    if (is == 0) Fdn = m_mq / 2.0;
    m_w4dn.setpart_ex(0, w, is_dn);
    m_fopr_w->mult_gm5(m_v4, m_w4dn);
    m_w4dn.daxpy(1.0, m_v4);
    v.addpart_ex(is, m_w4dn, 0, Fdn);
  }

  #pragma omp barrier

}

//====================================================================
//============================================================END=====
