/*!
        @file    $Id:: fopr_Domainwall.h #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru)
                 $LastChangedBy: matufuru $

        @date    $LastChangedDate:: 2014-02-04 23:44:44 #$

        @version $LastChangedRevision: 994 $
*/


#ifndef FOPR_DOMAINWALL_INCLUDED
#define FOPR_DOMAINWALL_INCLUDED

#include "fopr_Wilson.h"

#include "bridgeIO.h"
using Bridge::vout;

//! Domain-wall fermion operator.

/*!
    This class implements domain-wall fermion operator.
    At present, only the standard domain-wall fermion
    is ready.
                                     [24 Dec 2011 H.Matsufuru]
    (Coding history will be recovered from trac.)
    YAML is implemented.             [14 Nov 2012 Y.Namekawa]
    Multi-threaded.                  [09 Jan 2014 H.Matsufuru]
    Simplified for benchmark use     [07 Jun 2014 H.Matsufuru]
 */

class Fopr_Domainwall : public Fopr
{
 public:
  static const std::string class_name;

 private:
  // parameters common to overlap fermion
  int     m_NinF;          //!< on-site dof.
  int     m_Nvol;          //!< volume size.
  double  m_mq;            //!< quark mass.
  double  m_M0;            //!< domain-wall height.
  int     m_Ns;            //!< size of fifth-dimension.
  std::valarray<int> m_boundary;  //!< boundary conditions.
  std::string m_mode;      //!< mult mode.

  Fopr_Wilson *m_fopr_w;
  Field m_w1, m_w2, m_v1, m_v2;       //!< working vector.
  Field m_w4, m_v4, m_w4up, m_w4dn;   //!< working vector (4D).

 public:
  Fopr_Domainwall(Fopr_Wilson *fopr_w)
    : Fopr(), m_fopr_w(fopr_w)
  {
    init();
  }

 private:
  void init();

 public:
  void set_parameters(const double mq, const double M0, const int Ns,
                      const std::valarray<int> bc);

  void set_config(Field *U) { m_fopr_w->set_config(U); }

  void set_mode(std::string mode) { m_mode = mode; }

  std::string get_mode() const { return m_mode; }

  void mult(Field& v, const Field& f)
  {
    if (m_mode == "D") {
      D(v,f);
    } else if (m_mode == "DdagD") {
      DdagD(v,f);
    } else if (m_mode == "Ddag") {
      Ddag(v,f);
    } else if (m_mode == "H") {
      H(v,f);
    } else {
      vout.crucial(m_vl, "fopr_Domainwall: mode undefined.\n");
      abort();
    }
  }

  void mult_dag(Field& v, const Field& f)
  {
    if (m_mode == "D") {
      Ddag(v,f);
    } else if (m_mode == "DdagD") {
      DdagD(v,f);
    } else if (m_mode == "Ddag") {
      D(v,f);
    } else if (m_mode == "H") {
      H(v,f);
    } else {
      vout.crucial(m_vl, "fopr_Domainwall: mode undefined.\n");
      abort();
    }
  }

  const Field mult(const Field& f)
  {
    mult(m_v2,f);
    return m_v2;
  }

  const Field mult_dag(const Field& f)
  {
    mult_dag(m_v2,f);
    return m_v2;
  }

  void DdagD(Field&, const Field&);
  void D    (Field&, const Field&);
  void Ddag (Field&, const Field&);
  void H    (Field&, const Field&);
  void mult_gm5R(Field&, const Field&);

  //! chiral projection of 4D field.
  //!   ipm = 1: positive - (1+gm5)/2 multiplied
  //!        -1: negative - (1-gm5)/2
  void mult_chproj_4d(Field&, const Field&, int ipm);

  int field_nvol() { return m_Nvol; }
  int field_nin() {  return m_NinF; }
  int field_nex() {  return m_Ns; }

};

#endif
