/*!
        @file    $Id:: gaugeConfig.cpp #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru)
                 $LastChangedBy: matufuru $

        @date    $LastChangedDate:: 2014-06-03 23:17:04 #$

        @version $LastChangedRevision: 1057 $
*/

#include "gaugeConfig.h"

#include "commonParameters.h"
#include "communicator.h"
#include "bridgeIO.h"
using Bridge::vout;

#include "field.h"
#include "mat_SU_N.h"

#include "fieldIO.h"
#include "fieldIO_Text.h"
//#include "fieldIO_Binary.h"
//#include "fieldIO_Binary_Parallel.h"
//#include "fieldIO_Fortran.h"
//#include "fieldIO_LIME.h"
//#include "fieldIO_LIME_MPIIO.h"
#include "io_format.h"
#include "io_format_gauge.h"
//#include "randomNumbers_Mseries.h"

const std::string GaugeConfig::class_name = "GaugeConfig";

//====================================================================
void GaugeConfig::set_cold(Field *U)
{
  int Nc   = CommonParameters::Nc();
  int Nvol = CommonParameters::Nvol();
  int Ndim = CommonParameters::Ndim();

  Field_G  *up = (Field_G *)U;
  Mat_SU_N ut(Nc);

  ut.unit();

  for (int mu = 0; mu < Ndim; ++mu) {
    for (int site = 0; site < Nvol; ++site) {
      up->set_mat(site, mu, ut);
    }
  }

  vout.detailed("gauge config. was set to unity.\n");
}

//====================================================================
GaugeConfig::GaugeConfig(const string& type)
  : m_prefix(), m_postfix(), m_vl(CommonParameters::Vlevel()), m_fieldio()
{
  if (type == "Text") {
    m_fieldio = new FieldIO_Text(IO_Format::Gauge::ILDG);
    /*
  } else if (type == "Binary") {
    m_fieldio = new FieldIO_Binary(IO_Format::Gauge::ILDG);
  } else if (type == "Fortran_JLQCD") {
    m_fieldio = new FieldIO_Fortran(IO_Format::Gauge::JLQCD);
  } else if (type == "Fortran_ILDG") {
    m_fieldio = new FieldIO_Fortran(IO_Format::Gauge::ILDG);
    //  } else if (type == "ILDG") {
    //    m_fieldio = new FieldIO_LIME(IO_Format::Gauge::ILDG);
  } else if (type == "Binary_Parallel") {
    m_fieldio = new FieldIO_Binary_Parallel(IO_Format::Gauge::ILDG);
    //  } else if (type == "ILDG_Parallel") {
    //    m_fieldio = new FieldIO_LIME_MPIIO(IO_Format::Gauge::ILDG);
    */
  } else {
    vout.crucial("GaugeConfig: unsupported type \"%s\".\n", type.c_str());
    abort();
  }
}


//====================================================================
GaugeConfig::~GaugeConfig()
{
  if (m_fieldio) delete m_fieldio;
  m_fieldio = 0;
}


//====================================================================
string GaugeConfig::config_filename(const int confnum)
{
  const size_t filename_len = 256;
  char         buf[filename_len];

  snprintf(buf, filename_len, "%s%06d%s", m_prefix.c_str(), confnum, m_postfix.c_str());

  return string(buf);
}


//====================================================================
void GaugeConfig::read_file(Field *u, const string& filename)
{
  if (!m_fieldio) {
    vout.crucial("GaugeConfig::read_config(): FieldIO not set.");
    abort();
  }

  m_fieldio->read_file(u, filename);
}


//====================================================================
void GaugeConfig::write_file(Field *u, const string& filename)
{
  if (!m_fieldio) {
    vout.crucial("GaugeConfig::read_config(): FieldIO not set.");
    abort();
  }

  m_fieldio->write_file(u, filename);
}


//====================================================================
void GaugeConfig::read_config(Field *U, int confnum)
{
  return read_file(U, config_filename(confnum));
}


//====================================================================
void GaugeConfig::write_config(Field *U, int confnum)
{
  return write_file(U, config_filename(confnum));
}


//====================================================================
void GaugeConfig::read_file_4x4x4x8(Field *v, string filename)
{
  int nin_field = v->nin();
  int nex_field = v->nex();

  int Nc   = CommonParameters::Nc();
  int Ndim = CommonParameters::Ndim();
  int NinG = 2 * Nc * Nc;
  int NxF = 4;
  int NyF = 4;
  int NzF = 4;
  int NtF = 8;
  int LvolF = NxF *  NyF * NzF * NtF;
  Field v_in(NinG,LvolF,Ndim);

  int nin_file = NinG;
  int nex_file = Ndim;
  assert(nin_file == nin_field);
  assert(nex_file == nex_field);

  int Lvol = CommonParameters::Lvol();

  vout.detailed(m_vl, "%s: file format: nin=%d, nex=%d, Lvol=%d\n", __func__, nin_file, nex_file, Lvol);
  vout.detailed(m_vl, "%s: field format: nin=%d, nex=%d, Lvol=%d\n", __func__, nin_field, nex_field, v->nvol());

  // temporary field holding the whole space-time data.
  //  Field vtmp;
  Field vtmp(nin_field, LvolF, nex_field);

  if (Communicator::is_primary()) {
    vout.detailed(m_vl, "reading field data from %s\n", filename.c_str());

    //    vtmp.reset(nin_field, Lvol, nex_field);

    std::fstream config(filename.c_str(), std::ios::in);
    if (!config.is_open()) {
      vout.crucial(m_vl, "%s: file open error: %s may not exist.\n", __func__, filename.c_str());
      abort();
    }

    int    s, t;
    double val;
    for (int isite = 0; isite < LvolF; ++isite) {
     for (int j = 0; j < nex_file; ++j) {
       for (int i = 0; i < nin_file; ++i) {

          config >> val;

          if (!config.good()) {
            if (config.eof()) {
              // file short.
              vout.crucial(m_vl, "%s: file size too small.\n", __func__);
              abort();
            }
            if (config.fail()) {
              // invalid data.
              vout.crucial(m_vl, "%s: invalid data.\n", __func__);
              abort();
            }
            // other error.
            vout.crucial(m_vl, "%s: io error.\n", __func__);
            abort();
          }

	  //  m_format->file_to_field(s, t, i, j);
          vtmp.set(i, isite, j, val);
        }
      }
    }

    int dummy;
    config >> dummy;

    if (!config.eof()) {
      // file size mismatch
      vout.crucial(m_vl, "%s: file size larger than expected.\n", __func__);
      // warning only
    }

    config.close();
  }
  Communicator::sync();

  // vout.general(" U[0] = %18.12e\n",vtmp.cmp(0,0,0));
  // vout.general(" U[n] = %18.12e\n",vtmp.cmp(17,LvolF-1,3));

  //  FieldIO::deliver(v, &vtmp);
  vout.detailed(m_vl, "read successful\n");

  // copy read configuration to all the nodes.
  int count = NinG * LvolF * Ndim;
  Communicator::broadcast(count, (double*)vtmp.ptr(0), 0);
  //  broadcast(int count, double *data, int sender);

  Index_lex idx;
  Index_lex indexF(NxF, NyF, NzF, NtF);

  int Nx   = CommonParameters::Nx();
  int Ny   = CommonParameters::Ny();
  int Nz   = CommonParameters::Nz();
  int Nt   = CommonParameters::Nt();

  int ipex = Communicator::ipe(0);
  int ipey = Communicator::ipe(1);
  int ipez = Communicator::ipe(2);
  int ipet = Communicator::ipe(3);

  for (int j = 0; j < Ndim; ++j) {
   for (int t = 0; t < Nt; ++t) {
     int t2 = (t + ipet*Nt) % NtF;
    for (int z = 0; z < Nz; ++z) {
      int z2 = (z + ipez*Nz) % NzF;
     for (int y = 0; y < Ny; ++y) {
       int y2 = (y + ipey*Ny) % NyF;
      for (int x = 0; x < Nx; ++x) {
        int x2 = (x + ipex*Nx) % NxF;
        int lsite = idx.site(x, y, z, t);
        int gsite = indexF.site(x2, y2, z2, t2);

        for (int i = 0; i < NinG; ++i) {
           v->set(i, lsite, j, vtmp.cmp(i, gsite, j));
        }
      }
     }
    }
   }
  }

  // vout.general(" U[0] = %18.12e\n",v->cmp(0,0,0));
  // vout.general(" U[n] = %18.12e\n",v->cmp(17,Lvol-1,3));

}

//====================================================================
//============================================================END=====
